/*
 * Copyright 2015 Norbert Potocki (norbert.potocki@nort.pl)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.cfg4j.source.compose;

import static java.util.Objects.requireNonNull;

import org.cfg4j.source.ConfigurationSource;
import org.cfg4j.source.context.Environment;
import org.cfg4j.source.context.MissingEnvironmentException;

import java.util.Properties;

/**
 * Combines multiple {@link ConfigurationSource}s in a fallback chain. When one of the sources is not available
 * another one is used for providing configuration.
 */
public class FallbackConfigurationSource implements ConfigurationSource {

  private final ConfigurationSource[] sources;

  /**
   * Create a fallback chain of {@link ConfigurationSource}s
   *
   * @param sources configuration sources to use
   */
  public FallbackConfigurationSource(ConfigurationSource... sources) {
    this.sources = requireNonNull(sources);

    for (ConfigurationSource source : sources) {
      requireNonNull(source);
    }
  }

  /**
   * Get configuration set from this source in a form of {@link Properties}. Uses default environment. The configuration
   * set is a result of the first {@link #getConfiguration()} call to underlying sources that succeeds. Sources are
   * called in a provided order.
   *
   * @return configuration set for default environment taken from the first source that works
   * @throws IllegalStateException when unable to fetch configuration from any of the underlying sources
   */
  @Override
  public Properties getConfiguration() {

    for (ConfigurationSource source : sources) {
      try {
        return source.getConfiguration();
      } catch (IllegalStateException e) {
        // NOP
      }
    }

    throw new IllegalStateException();
  }

  /**
   * Get configuration set for a given {@code environment} from this source in a form of {@link Properties}.
   * Uses default environment. The configuration set is a result of the first {@link #getConfiguration()}
   * call to underlying sources that succeeds. Sources are called in a provided order.
   *
   * @param environment environment to use
   * @return configuration set for {@code environment} from the first source that works
   * @throws MissingEnvironmentException when requested environment couldn't be found in any of the underlying source
   * @throws IllegalStateException       when unable to fetch configuration from any of the underlying sources
   */
  @Override
  public Properties getConfiguration(Environment environment) {

    boolean allMissEnvironment = true;

    for (ConfigurationSource source : sources) {
      try {
        return source.getConfiguration(environment);
      } catch (MissingEnvironmentException e) {
        // NOP
      } catch (IllegalStateException e) {
        allMissEnvironment = false;
      }
    }

    if (allMissEnvironment) {
      throw new MissingEnvironmentException(environment.getName());
    }

    throw new IllegalStateException();
  }

  /**
   * Request configuration refresh. When this method returns configuration should be reloaded for at least
   * one underlying source. Tries to refresh all sources. Refresh failures are ignored (unless all sources fail to refresh).
   *
   * @throws IllegalStateException when unable to refresh all of the underlying sources
   */
  @Override
  public void refresh() {
    boolean atLeastOneSuccess = false;

    for (ConfigurationSource source : sources) {
      try {
        source.refresh();
        atLeastOneSuccess = true;
      } catch (IllegalStateException e) {
        // NOP
      }
    }

    if (!atLeastOneSuccess) {
      throw new IllegalStateException("Unable to refresh any of the underlying sources");
    }
  }

}
